from check import check_args,  check_args_for_class
from typing import (
    Any,
    Dict,
    List,
    NamedTuple,
    Tuple,
    TypeVar,
    Type,
    Literal,
    Union,
    Optional, Generic,Callable)


T = TypeVar('T', int, float, Dict[int, str])
P = TypeVar('P')


@check_args
def test_int(a: int):
    return True


@check_args
def test_float(a: float):
    return True


@check_args
def test_bool(a: bool):
    return True


@check_args
def test_string(a: str):
    return True


@check_args
def test_args(a, *args):
    return True


@check_args
def test_kwargs(a, **kwargs):
    return True


@check_args
def test_list(a: list):
    return True


@check_args
def test_List(a: List[Optional[int]]):
    return True


@check_args
def test_Any(a: Any):
    return True


@check_args_for_class
class A(object):
    def __init__(self):
        super(A, self).__init__()

    def func(self, a: T):
        return a


@check_args_for_class
class B(A):
    def __init__(self):
        super(B, self).__init__()

    def func(self, a: T, b: List[int]):
        return a


class C(object):
    def __init__(self):
        super(C, self).__init__()


class D(C, B):
    def __init__(self):
        super(D, self).__init__()


class Test_list(tuple):
    def __init__(self, *args):
        super(Test_list, self).__init__(*args)


@check_args
def test_extend(a: A):
    return True


@check_args
def test_Dict(a: Dict[str, List[int]]):
    return True


@check_args
def test_Union(a: Union[int, List[Optional[int]]]):
    return True


@check_args
def test_Tuple(a: Tuple[int or str]):
    return True


@check_args
def test_Literal(a: Literal['1', 2]):
    return True


@check_args
def test_TypeVar(a: T):
    return True


class Mapping(Generic[T, P]):
    @check_args
    def test(self, key: T) -> P:
        return False
@check_args



def test_call(a:Callable):
    return True

class Call_test():
    def __call__():
        return None

if __name__ == '__main__':
    test_int(1)
    try:
        test_int(2.2)
    except DetailedTypeError as e:
        print(e)
    test_float(2.2)
    try:
        test_float(1)
    except DetailedTypeError as e:
        print(e)
    test_bool(True)
    try:
        test_bool('False')
    except DetailedTypeError as e:
        print(e)
    test_string('aaa')
    try:
        test_bool(['a', 'a'])
    except DetailedTypeError as e:
        print(e)
    test_args(1, 2, 3)
    try:
        test_args(1, 2, b=5)
    except TypeError as e:
        print(e)
    test_kwargs(1, b=3, c=5)
    try:
        test_kwargs(1, {'a': 'B'})
    except TypeError as e:
        print(e)

    test_list([1, 2])
    try:
        test_list((1, 2))
    except DetailedTypeError as e:
        print(e)

    test_List([1, 2])
    try:
        test_List((1, 2))
    except DetailedTypeError as e:
        print(e)
    try:
        test_List(['1', 2])
    except DetailedTypeError as e:
        print(e)
    try:
        test_List({1: 2})
    except DetailedTypeError as e:
        print(e)
    try:
        test_List(set([1, 2]))
    except DetailedTypeError as e:
        print(e)
    test_Any(DeprecationWarning)
    test_Any(None)
    test_extend(A())
    test_extend(B())
    try:
        test_extend(C())
    except DetailedTypeError as e:
        print(e)
    test_Dict({'1': [1, 2]})
    try:
        test_Dict({1: 2, 3: 4})
    except DetailedTypeError as e:
        print(e)
    try:
        test_Dict([1, 2])
    except DetailedTypeError as e:
        print(e)
    test_Union(1)
    test_Union([1, None])
    try:
        test_Union(set([1, None]))
    except DetailedTypeError as e:
        print(e)
    test_Tuple((1, '2'))
    try:
        test_Tuple([1, '2'])
    except DetailedTypeError as e:
        print(e)
    test_Literal('1')
    try:
        test_Literal('2')
    except DetailedTypeError as e:
        print(e)
    test_TypeVar(1)
    test_TypeVar(1.1)
    test_TypeVar({1:'2'})
    try:
        test_TypeVar('2')
    except DetailedTypeError as e:
        print(e)
    try:
        test_TypeVar({1:1})
    except DetailedTypeError as e:
        print(e)
    a = A()
    b = B()
    a.func(1)
    try:
        a.func([1,2])
    except DetailedTypeError as e:
        print(e)
    b.func(1,[2])
    try:
        b.func(1,2)
    except DetailedTypeError as e:
        print(e)
    m = Mapping()
    m.test(1)
    try:
        m.test('111')
    except DetailedTypeError as e:
        print(e)
    test_call(Call_test())
    try:
        test_call(1)
    except DetailedTypeError as e:
        print(e)



from setuptools import setup
setup(
    name='parmscheck',
    version='1.0.7',
    author='Lfr',
    author_email='1733535832@qq.com',
    description='Decorator for type checking arguments ',
    license = 'Apache',
    packages=['parmscheck'],
    url='https://github.com/violetDelia/parmscheck',
    install_requires=[
    ],
)


from functools import wraps
import inspect
import typing
from typing import (
    Any,
    Dict,
    List,
    NamedTuple,
    Tuple,
    TypeVar,
    Type,
    Literal,
    Union)


def check_args(func):
    """
    The decorator that checks types at runtime in a function.
    """

    @wraps(func)
    def check(*args, **kwargs):
        sig = inspect.signature(func)
        binding = None
        try:
            binding = sig.bind(*args, **kwargs)
        except TypeError as te:
            raise DetailedTypeError(
                [Description(func.__name__, f'a TypeError is raised: {te}')])
        errors = []
        for name, value in binding.arguments.items():
            if not check_type(value, sig.parameters[name].annotation):
                errors.append(Description(
                    func.__name__, f'the type of variable "{name}" is not {sig.parameters[name].annotation}. Instead, it is {type(value)}'))
        if errors:
            raise DetailedTypeError(errors)
        return func(*args, **kwargs)
    return check


def check_args_for_class(cls):
    """
    The decorator that checks types at runtime for a class
    
    """
    for name, method in inspect.getmembers(cls):
        if (not inspect.ismethod(method) and not inspect.isfunction(method)) or inspect.isbuiltin(method):
            continue
        if method.__name__ == '__init__':
            continue
        setattr(cls, name, check_args(method))
    return cls


def check_type(obj: Any,
               candidate_type: Any,
               reltype: str = 'invariant'):
    '''
    Args:
        obj (Any): The value to check.
        candidate_type (Any): The type to check the object against.
        reltype (:obj:`str`, optional): Variance of the type.
    Returns:
        bool: True if the type is fine, False otherwise
    '''
    assert reltype in ['invariant', 'covariant',
                       'contravariant'], f' Variadic type {reltype} is unknown'
    # print(type(candidate_type))

    if candidate_type == inspect._empty:
        return True

    if type(candidate_type) == type and reltype in ['invariant']:
        return isinstance(obj, candidate_type)

    if type(candidate_type) == type and reltype in ['covariant']:
        return issubclass(obj.__class__, candidate_type)

    if type(candidate_type) == type and reltype in ['contravariant']:
        return issubclass(candidate_type, obj.__class__)

    if type(candidate_type) == TypeVar:
        if not candidate_type.__constraints__:
            return True
        if not (candidate_type.__covariant__ or candidate_type.__contravariant__):
            return any(check_type(obj, t) for t in candidate_type.__constraints__)

    # if type(candidate_type) == type(Type):
    #     return check_type(obj, candidate_type.__args__[0], reltype='covariant')

    if inspect.isclass(candidate_type) and reltype in ['invariant']:
        return isinstance(obj, candidate_type)

    if type(candidate_type) == typing._SpecialForm:
        if candidate_type._name == 'Any':
            return True
    elif type(candidate_type) == typing._GenericAlias:
        # print(candidate_type._name)
        if candidate_type._name == 'Union':
            return any(check_type(obj, t, reltype) for t in candidate_type.__args__)

        elif candidate_type._name == 'Dict':
            if not hasattr(obj, 'keys'):
                return False
            if not hasattr(obj, 'values'):
                return False
            if not hasattr(obj, 'items'):
                return False
            if type(obj) in [dict]:
                return all(check_type(k, candidate_type.__args__[0], reltype)
                           and check_type(v, candidate_type.__args__[1], reltype)
                           for (k, v) in obj.items())

        elif candidate_type._name == 'List':
            if not hasattr(obj, 'sort'):
                return False
            return all(check_type(o, candidate_type.__args__[0], reltype) for o in obj)

        elif candidate_type._name == 'Set':
            if not hasattr(obj, '__or__'):
                return False
            return all(check_type(o, candidate_type.__args__[0], reltype) for o in obj)

        elif candidate_type._name == 'Sized':
            if not hasattr(obj, '__len__'):
                return False
            return True

        elif candidate_type._name == 'Tuple':
            if hasattr(obj, '__setitem__'):
                return False
            if hasattr(obj, '__delitem__'):
                return False
            return all(check_type(o, t, reltype) for (o, t) in zip(obj, candidate_type.__args__))

        elif candidate_type._name == 'Sequence':
            if not hasattr(obj, '__len__'):
                return False
            return all(check_type(o, candidate_type.__args__[0], reltype) for o in obj)

        elif candidate_type._name is None:
            if candidate_type.__origin__ == typing.Generic:
                return all(check_type(o, t, reltype) for (o, t) in zip(obj, candidate_type.__args__))
            if candidate_type.__origin__ == Union:
                return any(check_type(obj, t, reltype) for t in candidate_type.__args__)
            if candidate_type.__origin__ == Literal:
                return obj in candidate_type.__args__

    elif type(candidate_type) == typing._VariadicGenericAlias:
        if candidate_type._name == 'Callable':
            if not hasattr(obj, '__call__'):
                return False
            return True
    #     return obj in candidate_type.__args__

    # elif type(candidate_type) == typing._UnionGenericAlias:
    #     return any(check_type(obj, t, reltype) for t in candidate_type.__args__)

    return False


class Description(NamedTuple):
    """Represents single type mismatch"""
    func_name: str
    message: str

    def __repr__(self) -> str:
        return f'in function {self.func_name}(),{self.message}'


class DetailedTypeError(TypeError):
    """Error for more detailed info about type mismatches"""
    issues = []

    def __init__(self, issues: List[Description]):
        self.issues = issues
        super().__init__(f'typing issues found:{issues}')

    def __str__(self):
        return '\n'.join(str(i) for i in self.issues)

    def __iter__(self):
        return (x for x in self.issues)

    def __len__(self):
        return len(self.issues)
